<?php
/**
 * Plugin main class.
 *
 * @package woocommerce-distance-rate-shipping
 */

use WooCommerce\DistanceRate\Blocks_Integration;
use WooCommerce\DistanceRate\Store_API_Extension;

/**
 * Distance rate plugin class.
 */
class WC_Distance_Rate {
	/**
	 * Plugin's version.
	 *
	 * @since 1.0.5
	 *
	 * @var string
	 */
	public $version;

	/**
	 * Constructor.
	 */
	public function __construct() {
		$this->version = WC_DISTANCE_RATE_VERSION;

		add_action( 'admin_init', array( $this, 'maybe_install' ), 5 );
		add_action( 'after_setup_theme', array( $this, 'load_plugin_textdomain' ) );
		add_filter( 'plugin_action_links_' . plugin_basename( WC_DISTANCE_RATE_FILE ), array( $this, 'plugin_action_links' ) );
		add_action( 'woocommerce_shipping_init', array( $this, 'shipping_init' ) );
		add_filter( 'woocommerce_shipping_methods', array( $this, 'shipping_methods' ) );
		add_action( 'admin_notices', array( $this, 'upgrade_notice' ) );
		add_action( 'wp_ajax_distance_rate_dismiss_upgrade_notice', array( $this, 'dismiss_upgrade_notice' ) );
		add_action( 'woocommerce_loaded', array( $this, 'load_post_wc_class' ) );
		add_action( 'before_woocommerce_init', array( $this, 'declare_hpos_compatibility' ) );
		add_action( 'admin_enqueue_scripts', array( $this, 'admin_enqueue_scripts' ) );
		add_action( 'woocommerce_blocks_loaded', array( $this, 'register_blocks_integration' ) );
		add_action( 'woocommerce_blocks_loaded', array( $this, 'extend_store_api' ) );
	}

	/**
	 * Loads any class that needs to check for WC loaded.
	 *
	 * @since 1.0.9
	 */
	public function load_post_wc_class() {
		require_once __DIR__ . '/class-wc-distance-rate-privacy.php';
	}

	/**
	 * Localisation.
	 */
	public function load_plugin_textdomain() {
		load_plugin_textdomain( 'woocommerce-distance-rate-shipping', false, plugin_basename( WC_DISTANCE_RATE_ABSPATH ) . '/languages/' );
	}

	/**
	 * Add plugin action links to the plugins page.
	 *
	 * @param  array $links Links.
	 * @return array
	 */
	public function plugin_action_links( $links ) {
		$plugin_links = array(
			'<a href="' . admin_url( ( function_exists( 'WC' ) ? 'admin.php?page=wc-settings&tab=shipping&section=wc_shipping_distance_rate' : 'admin.php?page=woocommerce_settings&tab=shipping&section=WC_Distance_Rate' ) ) . '">' . __( 'Settings', 'woocommerce-distance-rate-shipping' ) . '</a>',
			'<a href="https://support.woocommerce.com/">' . __( 'Support', 'woocommerce-distance-rate-shipping' ) . '</a>',
			'<a href="https://docs.woocommerce.com/document/woocommerce-distance-rate-shipping/">' . __( 'Docs', 'woocommerce-distance-rate-shipping' ) . '</a>',
		);

		return array_merge( $plugin_links, $links );
	}

	/**
	 * Declaring HPOS compatibility.
	 */
	public function declare_hpos_compatibility() {
		if ( class_exists( '\Automattic\WooCommerce\Utilities\FeaturesUtil' ) ) {
			\Automattic\WooCommerce\Utilities\FeaturesUtil::declare_compatibility( 'custom_order_tables', 'woocommerce-distance-rate-shipping/woocommerce-distance-rate-shipping.php', true );
		}
	}

	/**
	 * Load our shipping class.
	 */
	public function shipping_init() {
			include_once __DIR__ . '/class-wc-shipping-distance-rate.php';
	}

	/**
	 * Add our shipping method to woocommerce.
	 *
	 * @param  array $methods Shipping methods.
	 * @return array
	 */
	public function shipping_methods( $methods ) {
			$methods['distance_rate'] = 'WC_Shipping_Distance_Rate';
		return $methods;
	}

	/**
	 * Checks the plugin version.
	 *
	 * @since 1.0.5
	 * @version 1.0.5
	 * @return bool
	 */
	public function maybe_install() {
		// Only need to do this for versions less than 1.0.5 to migrate
		// settings to shipping zone instance.
		$doing_ajax = defined( 'DOING_AJAX' ) && DOING_AJAX;
		if ( ! $doing_ajax
				&& ! defined( 'IFRAME_REQUEST' )
				&& version_compare( get_option( 'wc_distance_rate_version' ), '1.0.5', '<' ) ) {

			$this->install();

		}

		return true;
	}

	/**
	 * Admin styles + scripts.
	 */
	public function admin_enqueue_scripts() {
		$suffix = ( defined( 'SCRIPT_DEBUG' ) && SCRIPT_DEBUG ) ? '' : '.min';

		wp_register_script( 'wc_shipping_distance_rates_upgrade', WC_DISTANCE_RATE_PLUGIN_URL . 'assets/js/upgrade-notice' . $suffix . '.js', array( 'jquery', 'wp-util' ), WC_DISTANCE_RATE_VERSION, true );

		wp_enqueue_style( 'wc_shipping_distance_rates_zone', WC_DISTANCE_RATE_PLUGIN_URL . 'assets/css/shipping-zone.css', array(), WC_DISTANCE_RATE_VERSION );
		wp_register_script( 'wc_shipping_distance_rates_zone', WC_DISTANCE_RATE_PLUGIN_URL . 'assets/js/shipping-zone' . $suffix . '.js', array( 'jquery', 'wp-util' ), WC_DISTANCE_RATE_VERSION, true );
		wp_localize_script(
			'wc_shipping_distance_rates_zone',
			'wc_distance_rate_params',
			array(
				'id'   => 'distance_rate',
				'i18n' => array(
					'distance'          => __( 'Distance', 'woocommerce-distance-rate-shipping' ),
					'total_travel_time' => __( 'Total Travel Time', 'woocommerce-distance-rate-shipping' ),
					'weight'            => __( 'Weight', 'woocommerce-distance-rate-shipping' ),
					'order_total'       => __( 'Order Total', 'woocommerce-distance-rate-shipping' ),
					'quantity'          => __( 'Quantity', 'woocommerce-distance-rate-shipping' ),
					'and'               => __( 'and', 'woocommerce-distance-rate-shipping' ),
					'or'                => __( 'or', 'woocommerce-distance-rate-shipping' ),
					'add_rule_group'    => __( 'Add rule group', 'woocommerce-distance-rate-shipping' ),
					'distance'          => __( 'Distance', 'woocommerce-distance-rate-shipping' ),
					'travel_time'       => __( 'Travel Time', 'woocommerce-distance-rate-shipping' ),
					'del_selected_rule' => __( 'Delete the selected rule?', 'woocommerce-distance-rate-shipping' ),
					'format_price'      => wc_format_localized_price( 0 ),
					'abort_text'        => __( 'Abort text', 'woocommerce-distance-rate-shipping' ),
				),
			)
		);
	}

	/**
	 * Update/migration script.
	 *
	 * @since 1.0.5
	 * @version 1.0.5
	 */
	public function install() {
		// Get all saved settings and cache it.
		$distance_rate_settings = get_option( 'woocommerce_distance_rate_settings', false );

		// Rules are stored in separate tables.
		$distance_rate_rules = get_option( 'woocommerce_distance_rate_rules', array() );

		// Settings exists.
		if ( $distance_rate_settings ) {
			global $wpdb;

			// Unset un-needed settings.
			unset( $distance_rate_settings['enabled'] );
			unset( $distance_rate_settings['availability'] );
			unset( $distance_rate_settings['countries'] );

			// Merge rules into settings.
			$distance_rate_settings['rules'] = $distance_rate_rules;

			// First add it to the "rest of the world" zone when no distance rate
			// instance.
			if ( ! $this->is_zone_has_distance_rate( 0 ) ) {
				$wpdb->query( $wpdb->prepare( "INSERT INTO {$wpdb->prefix}woocommerce_shipping_zone_methods ( zone_id, method_id, method_order, is_enabled ) VALUES ( %d, %s, %d, %d )", 0, 'distance_rate', 1, 1 ) );
				// Add settings to the newly created instance to options table.
				$instance = $wpdb->insert_id;
				add_option( 'woocommerce_distance_rate_' . $instance . '_settings', $distance_rate_settings );
			}

			update_option( 'woocommerce_distance_rate_show_upgrade_notice', 'yes' );
		}

		update_option( 'wc_distance_rate_version', $this->version );
	}

	/**
	 * Show the user a notice for plugin updates.
	 *
	 * @since 1.0.5
	 */
	public function upgrade_notice() {
		$show_notice = get_option( 'woocommerce_distance_rate_show_upgrade_notice' );

		if ( 'yes' !== $show_notice ) {
			return;
		}

		wp_enqueue_script( 'wc_shipping_distance_rates_upgrade' );

		$query_args = array(
			'page' => 'wc-settings',
			'tab'  => 'shipping',
		);

		$zones_admin_url = add_query_arg( $query_args, get_admin_url() . 'admin.php' );
		?>
		<div class="notice notice-success is-dismissible wc-distance-rate-notice">
			<p>
			<?php
				/* translators: %1s: Shipping zones link start %2s: Link end */
				printf( esc_html__( 'Distance Rate now supports shipping zones. The zone settings were added to a new Distance Rate method on the "Rest of the World" Zone. See the zones %1$shere%2$s ', 'woocommerce-distance-rate-shipping' ), '<a href="' . esc_url( $zones_admin_url ) . '">', '</a>' );
			?>
			</p>
		</div>
		<?php
	}

	/**
	 * Turn of the dismisable upgrade notice.
	 *
	 * @since 1.0.5
	 */
	public function dismiss_upgrade_notice() {
		update_option( 'woocommerce_distance_rate_show_upgrade_notice', 'no' );
	}

	/**
	 * Helper method to check whether given zone_id has distance rate method instance.
	 *
	 * @since 1.0.5
	 *
	 * @param int $zone_id Zone ID.
	 *
	 * @return bool True if given zone_id has distance rate method instance
	 */
	public function is_zone_has_distance_rate( $zone_id ) {
		global $wpdb;

		return (int) $wpdb->get_var( $wpdb->prepare( "SELECT COUNT(instance_id) FROM {$wpdb->prefix}woocommerce_shipping_zone_methods WHERE method_id = 'distance_rate' AND zone_id = %d", $zone_id ) ) > 0;
	}

	/**
	 * Register blocks integration.
	 */
	public function register_blocks_integration() {
		require_once WC_DISTANCE_RATE_ABSPATH . 'includes/class-blocks-integration.php';
		add_action(
			'woocommerce_blocks_cart_block_registration',
			function ( $integration_registry ) {
				$integration_registry->register( new Blocks_Integration() );
			}
		);
		add_action(
			'woocommerce_blocks_checkout_block_registration',
			function ( $integration_registry ) {
				$integration_registry->register( new Blocks_Integration() );
			}
		);
	}

	/**
	 * Extend the store API.
	 */
	public function extend_store_api() {
		require_once WC_DISTANCE_RATE_ABSPATH . 'includes/class-store-api-extension.php';
		Store_API_Extension::init();
	}
}
